#ifndef _UDM_MUTEX_H
#define _UDM_MUTEX_H

#ifdef HAVE_PTHREAD
extern __INDLIB__ int InitMutexes();
extern __INDLIB__ int DestroyMutexes();
#endif


/* MUTEX stuff to lock dangerous places in multi-threaded mode */

#if (WIN32|WINNT)

#ifdef HAVE_PTHREAD
extern CRITICAL_SECTION robots_mutex;
extern CRITICAL_SECTION hrefs_mutex;
extern CRITICAL_SECTION docs_mutex;
extern CRITICAL_SECTION target_mutex;
extern CRITICAL_SECTION indexed_mutex;
extern CRITICAL_SECTION query_mutex;
#define LOCK_ROBOTS()		EnterCriticalSection(&robots_mutex)
#define UNLOCK_ROBOTS()		LeaveCriticalSection(&robots_mutex)
#define LOCK_HREFS(x)		EnterCriticalSection(&hrefs_mutex)
#define UNLOCK_HREFS(x)		LeaveCriticalSection(&hrefs_mutex)
#define LOCK_DOCS(x)		EnterCriticalSection(&docs_mutex)
#define UNLOCK_DOCS(x)		LeaveCriticalSection(&docs_mutex)
#define LOCK_TARGET(x)		EnterCriticalSection(&target_mutex)
#define UNLOCK_TARGET(x)	LeaveCriticalSection(&target_mutex)
#define LOCK_INDEXED()		EnterCriticalSection(&indexed_mutex)
#define UNLOCK_INDEXED()	LeaveCriticalSection(&indexed_mutex)
#define LOCK_MAIN()		EnterCriticalSection(&main_mutex)
#define UNLOCK_MAIN()		LeaveCriticalSection(&main_mutex)
#define LOCK_QUERY()		EnterCriticalSection(&query_mutex)
#define UNLOCK_QUERY()		LeaveCriticalSection(&query_mutex)
#define LOCK_HOST()
#define UNLOCK_HOST()

#else				/* If no threads for WIN */

#define LOCK_ROBOTS()
#define UNLOCK_ROBOTS()
#define LOCK_HREFS(x)
#define UNLOCK_HREFS(x)
#define LOCK_DOCS(x)
#define UNLOCK_DOCS(x)
#define LOCK_TARGET(x)
#define UNLOCK_TARGET(x)
#define LOCK_INDEXED()
#define UNLOCK_INDEXED()
#define LOCK_MAIN()
#define UNLOCK_MAIN()
#define LOCK_HOST()
#define UNLOCK_HOST()
#define LOCK_QUERY()
#define UNLOCK_QUERY()
#endif

#else				/* UNIX stuff */


#ifdef HAVE_PTHREAD
#include <pthread.h>
extern pthread_mutex_t robots_mutex;
extern pthread_mutex_t hrefs_mutex;
extern pthread_mutex_t docs_mutex;
extern pthread_mutex_t target_mutex;
extern pthread_mutex_t indexed_mutex;
extern pthread_mutex_t query_mutex;
extern pthread_mutex_t host_mutex;
#define LOCK_ROBOTS()		pthread_mutex_lock(&robots_mutex)
#define UNLOCK_ROBOTS()		pthread_mutex_unlock(&robots_mutex)
#define LOCK_HREFS(x)		pthread_mutex_lock(&hrefs_mutex)
#define UNLOCK_HREFS(x)		pthread_mutex_unlock(&hrefs_mutex)
#define LOCK_DOCS(x)		pthread_mutex_lock(&docs_mutex)
#define UNLOCK_DOCS(x)		pthread_mutex_unlock(&docs_mutex)
#define LOCK_TARGET(x)		pthread_mutex_lock(&target_mutex)
#define UNLOCK_TARGET(x)	pthread_mutex_unlock(&target_mutex)
#define LOCK_INDEXED()		pthread_mutex_lock(&indexed_mutex)
#define UNLOCK_INDEXED()	pthread_mutex_unlock(&indexed_mutex)
#define LOCK_MAIN()		pthread_mutex_lock(&main_mutex)
#define UNLOCK_MAIN()		pthread_mutex_unlock(&main_mutex)
#define LOCK_HOST()		pthread_mutex_lock(&host_mutex)
#define UNLOCK_HOST()		pthread_mutex_unlock(&host_mutex)
#define LOCK_QUERY()		pthread_mutex_lock(&query_mutex)
#define UNLOCK_QUERY()		pthread_mutex_unlock(&query_mutex)

#else				/* if no threads for UNIX */

#define LOCK_ROBOTS()
#define UNLOCK_ROBOTS()
#define LOCK_HREFS(x)
#define UNLOCK_HREFS(x)
#define LOCK_DOCS(x)
#define UNLOCK_DOCS(x)
#define LOCK_TARGET(x)
#define UNLOCK_TARGET(x)
#define LOCK_INDEXED()
#define UNLOCK_INDEXED()
#define LOCK_MAIN()
#define UNLOCK_MAIN()
#define LOCK_HOST()
#define UNLOCK_HOST()
#define LOCK_QUERY()
#define UNLOCK_QUERY()
#endif				/* HAVE_PTHREADS for UNIX */

#endif				

#endif				/* WIN32|WINNT */

