#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"
#include "udm_db.h"
#include "udm_hrefs.h"
#include "udm_utils.h"
#include "udm_mutex.h"


/* Max URLs in cache */
#define MAXHSIZE	1023
#define HSIZE           256


static	UDM_HREF* Href=NULL;
static	int mhrefs=0;
int nhrefs=0;


__INDLIB__ int AddHref(UDM_INDEXER *Indexer,char *href,int referrer,int hops,int stored){
int i;

	if(!href[0])return(0);
	if(strlen(href)>UDM_URLSIZE)return(0);
	if(Indexer)LOCK_HREFS(Indexer);
	for(i=0;i<nhrefs;i++)
		if(!strcmp(Href[i].href,href)){
			Href[i].stored|=stored;
			if(Indexer)UNLOCK_HREFS(Indexer);
			return(0);
		}
	if(nhrefs>=mhrefs){
		if(mhrefs){
			mhrefs+=HSIZE;
			Href=(UDM_HREF *)realloc(Href,mhrefs*sizeof(UDM_HREF));
		}else{
			mhrefs=HSIZE;
			Href=(UDM_HREF *)malloc(mhrefs*sizeof(UDM_HREF));
		}
	}
	Href[nhrefs].href=strdup(href);
	Href[nhrefs].referrer=referrer;
	Href[nhrefs].hops=hops;
	Href[nhrefs].stored=stored;
	nhrefs++;
	if(Indexer)UNLOCK_HREFS(Indexer);
	return(1);
}
void FreeHrefs(){
int i;
	for(i=0;i<nhrefs;i++)free(Href[i].href);
		nhrefs=0;mhrefs=0;
	UDM_FREE(Href);
}
int StoreHrefs(UDM_INDEXER * Indexer){
int i;
int added=0;

	LOCK_HREFS(Indexer);
	for(i=0;i<nhrefs;i++){
		if(!(Href[i].stored)){
			char * msg_id="";

			added++;

			/* see if it is a message */
			if(strchr(Href[i].href,'@')){
				char *pos;
				pos = strrchr(Href[i].href,'/');
				msg_id = malloc( strlen(pos) +1);
				strcpy(msg_id, (pos + 1));
			}
			// now add the url with its message id
			AddURL(Indexer,Href[i].href,Href[i].referrer,Href[i].hops,msg_id);

			if(DBErrorCode(Indexer->db)){
				UNLOCK_HREFS(Indexer);
				return(added);
			}
			// free msg_id if alloc'd
			if(strlen(msg_id))
				free(msg_id);
			Href[i].stored=1;
		}
	}

	/* We should not free URL list with onw database */
	/* to avoid double indexing of the same document */

#ifdef HAVE_SQL
	if(nhrefs>MAXHSIZE)
		FreeHrefs();
#endif

	UNLOCK_HREFS(Indexer);
	return(added);
}
