#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <regex.h>
#include <signal.h>
#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif
#if (WIN32|WINNT)
#include <time.h>
#else
#include <sys/time.h>
#endif

#include "udm_common.h"
#include "udm_log.h"
#include "udm_conf.h"
#include "udm_indexer.h"
#include "udm_robots.h"
#include "udm_db.h"
#include "udm_parseurl.h"
#include "udm_charset.h"
#include "udm_parser.h"
#include "udm_spell.h"
#include "udm_proto.h"
#include "udm_hrefs.h"
#include "udm_utils.h"
#include "udm_mutex.h"
#include "udm_md5.h"

#ifdef NEWS_EXT
#include "udm_parsedate.h"
#endif

#define MD5_DIGEST(data, size, digest) md5_buffer_str(data, size, digest)

static int have_targets=1;
static char *url_file_name=NULL;


/******************** URL number *******************************/
static  int  MaxURLNumber=-1;
static  int  CurURLNumber=0;
void IncURLNumber(){
	LOCK_INDEXED();
	CurURLNumber++;
	UNLOCK_INDEXED();
}
int GetIndexedNumber(){
int result;
	LOCK_INDEXED();
	result=CurURLNumber;
	UNLOCK_INDEXED();
	return(result);
}
__INDLIB__ void SetMaxURLNumber(int n){
	LOCK_INDEXED();
	MaxURLNumber = n;
	UNLOCK_INDEXED();
}


/********************  Words stuff *****************************/

#define RESORT_WORDS	256
#define WSIZE		1024

static int cmpwords(const void * v1, const void * v2){
	return(strcmp(((UDM_WORD*)v1)->word,((UDM_WORD*)v2)->word));
}
static int AddOneWord(UDM_INDEXER *Indexer,UDM_SERVER *Server,char *word,int where,int checkstop){
int i;
char *s;
UDM_STOPWORD * stop;
int l,r,c,res,wlen;

	wlen=strlen(word);
	if((wlen>Server->max_word_length)||(wlen<Server->min_word_length))
		return(0);
	s=strdup(word);
	if(strlen(s)>UDM_MAXWORDSIZE)s[UDM_MAXWORDSIZE]=0;
	tolower_string(s,LCindex);
	if(checkstop){
		/* Stopwords checking and language guesser */
		if((stop=IsStopWord(s))){
			int i;
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"stop: '%s'-'%s'", stop->word, stop->lang);
			/* Add language */
			if(!strcmp(stop->lang,""))
				Indexer->lang[0].count++;
			else
			for(i=1;i<UDM_LANGPERDOC;i++){
				if(!strcmp(Indexer->lang[i].lang,stop->lang)){
					Indexer->lang[i].count++;
					break;
				}
				/* if current lang is free */
				if(!strcmp(Indexer->lang[i].lang,"")){
					strcpy(Indexer->lang[i].lang,stop->lang);
					Indexer->lang[i].count++;
					break;
				}
			}
			free(s);
			return(0);
		}
	}
	/* Find current word in sorted part of word list */
	l=0;r=Indexer->swords-1;
	while(l<=r){
		c=(l+r)/2;
		res=strcmp(Indexer->Word[c].word,s);
		if(res==0){
			Indexer->Word[c].count|=where;
			free(s);
			return(0);
		}
		if(res<0)
			l=c+1;
		else
			r=c-1;
	}
	/* Find in unsorted part */
	for(i=Indexer->swords;i<Indexer->nwords;i++){
		if(!strcmp(Indexer->Word[i].word,s)){
			Indexer->Word[i].count|=where;
			free(s);return(0);
		}
	}
	/* Add new word */
	if(Indexer->nwords>=Indexer->mwords){
		if(Indexer->mwords){
			Indexer->mwords+=WSIZE;
			Indexer->Word=(UDM_WORD *)realloc(Indexer->Word,Indexer->mwords*sizeof(UDM_WORD));
		}else{
			Indexer->mwords=WSIZE;
			Indexer->Word=(UDM_WORD *)malloc(Indexer->mwords*sizeof(UDM_WORD));
		}
	}
	Indexer->Word[Indexer->nwords].word=s;
	Indexer->Word[Indexer->nwords].count=where;
	Indexer->nwords++;

	/* Sort unsorted part */
	if((Indexer->nwords-Indexer->swords)>RESORT_WORDS){
		qsort(Indexer->Word,Indexer->nwords,sizeof(UDM_WORD),cmpwords);
		Indexer->swords=Indexer->nwords;
	}
	return(0);
}

/* This function adds a normalized word form(s) into list using Ispell */
static int AddWord(UDM_INDEXER *Indexer,UDM_SERVER *Server,char *word,int where,int checkstop){
char 	** forms, ** saveforms;
int	have_digit=0;
int	have_alpha=0;

	if(Server->number_factor==0||Server->alnum_factor==0){
		char *s;
		s=word;
		while(*s){
			if(isdigit(*s))have_digit=1;
			if(isalpha(*s))have_alpha=1;
			if(have_digit&&have_alpha)break;
			s++;
		}
		if(have_digit){
			if(have_alpha){
				if(!Server->alnum_factor)return(0);
			}else{
				if(!Server->number_factor)return(0);
			}
		}
	}
	tolower_string(word,LCindex);
	if((saveforms=forms=NormalizeWord(word))){
		/* Add all NORMAL forms of the word */
		while(*forms){
			/* Add only if correct words are allowed */
			if(Server->correct_factor){
				AddOneWord(Indexer,Server,*forms,where,checkstop);
			}
			free(*forms);
			forms++;
		}
		free(saveforms);
	}else{
		/* If NORMAL forms has not been found  */
		/*   then we will add the word itself  */
		/* Do it only when incorrect words are */
		/* allowed by configuration            */
		if(Server->incorrect_factor)
			AddOneWord(Indexer,Server,word,where,checkstop);
	}
	return(0);
}

static int FreeWords(UDM_INDEXER* Indexer) {
	int i;
	for(i=0;i<Indexer->nwords;i++)
		free(Indexer->Word[i].word);
	Indexer->nwords=0;
	Indexer->swords=0;
	return(0);
}


/**************************** Built-in Parsers ***************************/

static int ParseText(UDM_INDEXER * Indexer,UDM_SERVER * Server,char *content,char *separator,int weight,int check_stopwords){
char *s,*lt;
	if(weight&&content&&separator){
		s=gettoken(content,separator,&lt);
		while(s){
			AddWord(Indexer,Server,s,weight,check_stopwords);
			s=gettoken(NULL,separator,&lt);
		}
	}return(0);
}


/* HTML parser states */
#define HTML_UNK	0
#define HTML_TAG	1
#define HTML_TXT	2
#define HTML_COM	3

static int ParseHtml(
UDM_INDEXER * Indexer,
UDM_SERVER * CurSrv,
char *content,
char *separator,
UDM_URL * curURL,
UDM_DOCUMENT * doc,
int index,int follow,
char *text,
char *keywords,
char *descript,
char *title){

int len,inbody,inscript,intitle,i,j,state,Method;
char sch=0,*s=0,*e=0,*tmp=0,*href=0,*newschema=0,*lt;
UDM_URL newURL;UDM_TAG tag;
UDM_FILTER *MatchedFilter=NULL;
char str[UDMSTRSIZ]="";
char str1[UDMSTRSIZ]="";

	inbody=inscript=intitle=0;state=HTML_UNK;s=content;
	while(*s){
		switch(state){
		case HTML_UNK: /* yet unknown */
			if(!UDM_STRNCMP(s,"<!--"))
				state=HTML_COM;
			else
			if(*s=='<')
				state=HTML_TAG;
			else
				state=HTML_TXT;
			break;


		case HTML_TAG: /* tag */
			e=strchr(s,'>');
			if(e)len=e-s+1;
			else len=strlen(s);
			tmp=(char*)malloc(len+1);
			strncpy(tmp,s,len);
			tmp[len]=0;
			parsetag(&tag,tmp);

			if(!strcasecmp(tag.name,"body"))inbody=1;
			else
			if(!strcasecmp(tag.name,"/body"))inbody=0;
			else
			if(!strcasecmp(tag.name,"script"))inscript=1;
			else
			if(!strcasecmp(tag.name,"/script"))inscript=0;
			else
			if(!strcasecmp(tag.name,"title"))intitle=1;
			else
			if(!strcasecmp(tag.name,"/title"))intitle=0;
			else
			if((!strcasecmp(tag.name,"a"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"area"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"link"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"frame"))&&(tag.src))
				href=strdup(tag.src);
			else
			if((!strcasecmp(tag.name,"img"))&&(tag.src))
				href=strdup(tag.src);
			else
			if((!strcasecmp(tag.name,"base"))&&(tag.href)){
				/* <BASE HREF="xxx"> stuff            */
				/* Check that URL is properly formed  */
				/* baseURL is just temporary variable */
				/* If parsing is fail we'll use old   */
				/* base href, passed via curURL       */
				UDM_URL baseURL;
				if(!parseurl(&baseURL,tag.href)){
					freeurl(&baseURL);
					freeurl(curURL);
					parseurl(curURL,tag.href);
				}
			}else
			if((!strcasecmp(tag.name,"meta"))&&(tag.equiv)&&(tag.content)){
#ifndef USE_CHARSET_GUESSER
				if((!strcasecmp(tag.equiv,"Content-Type"))&&(!Indexer->charset)){
					char *p;
					if((p=strstr(tag.content,"charset="))){
						Indexer->charset = p + 8;
						recode_string(content,getcharset(Indexer->charset),LCindex);
					}
				}else
#endif
				if(!strcasecmp(tag.equiv,"refresh")){
					char *u;
					if((u=strstr(tag.content,"URL=")))
						href=strdup(u+4);
				}else
				if(!strcasecmp(tag.equiv,"keywords")){
					strncpy(keywords,tag.content,UDM_MAXKEYWORDSIZE-1);
					keywords[UDM_MAXKEYWORDSIZE-1]=0;
					if(index&&CurSrv->keywordweight){
						strcpy(str,keywords);
						ParseText(Indexer,CurSrv,str,separator,
							CurSrv->keywordweight,1);
					}
				}else
				if(!strcasecmp(tag.equiv,"description")){
					strncpy(descript,tag.content,UDM_MAXDESCSIZE-1);
					descript[UDM_MAXDESCSIZE-1]=0;
					if(index&&CurSrv->descweight){
						strcpy(str,descript);
						ParseText(Indexer,CurSrv,str,separator,
							CurSrv->descweight,1);
					}
				}else
				if(!strcasecmp(tag.equiv,"robots")){
					if(CurSrv->userobots){
					e=gettoken(tag.content,separator,&lt);
					while(e){
						if(!strcasecmp(e,"ALL")){
							follow=1;index=1;
						}else
						if(!strcasecmp(e,"NONE")){
							follow=0;index=0;
						}else
						if(!strcasecmp(e,"NOINDEX"))
							index=0;
						else
						if(!strcasecmp(e,"NOFOLLOW"))
							follow=0;
						else
						if(!strcasecmp(e,"INDEX"))
							index=1;
						else
						if(!strcasecmp(e,"FOLLOW"))
							follow=1;
						e=gettoken(NULL,separator,&lt);
					}}
				}
			}
			if((href)&&(follow)&&(CurSrv->gfollow)){
				if(doc->hops>=CurSrv->maxhops){
					udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : too many hops: %d",href,doc->hops);
				}else
				if(!parseurl(&newURL,href)){
					if(newURL.schema[0])newschema=newURL.schema;
					else	newschema=curURL->schema;
					if(!strcmp(newschema,"file")||!strcmp(newschema,"htdb")){
						sprintf(str,"%s:%s%s",newschema,newURL.path[0]?newURL.path:curURL->path,newURL.filename);
					}else{
						sprintf(str,"%s://%s%s%s",
						newURL.schema[0]?newURL.schema:curURL->schema,
						newURL.hostinfo[0]?newURL.hostinfo:curURL->hostinfo,
						newURL.path[0]?newURL.path:curURL->path,newURL.filename);
					}
					remove2dot(str);
					if(!UDM_STRNCMP(str,"ftp://")&&(strstr(str,";type=")))
						*(strstr(str,";type"))=0;

					if((MatchedFilter=FindFilter(str)))
						Method=MatchedFilter->filter_type;
					else
						Method=UDM_ALLOW;

					if((Method==UDM_DISALLOW)){
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : '%s'",href,FilterComStr(MatchedFilter));
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Full URL \"%s\"",str);
					}else
					if((FindRobots(Indexer,str)>=0)&&(CurSrv->userobots)){
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : robots",href);
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Full URL \"%s\"",str);
					}else
					if(!FindServer(str)){
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : no Server",href);
					}else{
						int add=1;
						/* compare hostinfo in some cases */
						if((!CurSrv->delete_no_server)&&(!CurSrv->outside)){
							if(newURL.hostinfo[0])
								add=!strcmp(curURL->hostinfo,newURL.hostinfo);
						}
						if(add){
							/* Add robots.txt for HTTP schema */
							/* When FollowOutside or DeleteNoServer no */
							if((!strcmp(newURL.schema,"http"))&&(CurSrv->userobots)&&
							((CurSrv->outside)||(!CurSrv->delete_no_server))){
								sprintf(str1,"%s://%s/%s",newURL.schema,newURL.hostinfo,"robots.txt");
								if(AddHref(NULL,str1,0,0,0)){
									LOCK_TARGET(Indexer);
									have_targets=1;
									UNLOCK_TARGET(Indexer);
								}
							}
							if(AddHref(NULL,str,doc->url_id,doc->hops+1,0)){
								LOCK_TARGET(Indexer);
								have_targets=1;
								UNLOCK_TARGET(Indexer);
							}
						}
					}
					freeurl(&newURL);
				}
				UDM_FREE(href);
			}
			freetag(&tag);free(tmp);
			s=s+len; state=HTML_UNK;
			break;

		case HTML_COM: /* comment */
			e=strstr(s,"-->");
			if(e)len=e-s+3;
			else len=strlen(s);

			/* This is for debug pusposes
			tmp=(char*)malloc(len+1);
			strncpy(tmp,s,len);tmp[len]=0;
			printf("com: '%s'\n",tmp);
			*/

			s=s+len;
			state=HTML_UNK;
			break;

		case HTML_TXT: /* text */
			e=strchr(s,'<');

			/* Special case when script body is not commented: */
			/* <script> x="<"; </script> */

			/* We should find </script> in this case: */
			while((inscript)&&(e)&&(UDM_STRNCASECMP(e,"</script")))
				e=strchr(e+1,'<');

			if(e) len=e-s;
			else  len=strlen(s);
			tmp=(char*)malloc(2*len+6);
			e=tmp;j=0;*e=0;
			for(i=0;i<len;i++){
				switch(s[i]){
				case ' ' : case '\t':
				case '\n': case '\r':
					j++; break;

				case '&': /* parse specials */
					if(j){*e=' ';e++;j=0;}
					i++;
					if((sch=SgmlToChar(s+i,LCindex))){
						if(sch==' ')j++;
						*e=sch;e++;*e=0;
						while((i<len)&&(isalnum(*(s+i+1))))
							i++;
						if(s[i+1]==';')i++;
					}else{
						*e='?';e++;*e=0;
					}
					break;
				default:
					if(j){*e=' ';e++;j=0;}
					*e=(s[i]=='\'')?'`':s[i];
					e++; *e=0;
					break;
				}
			}
			if((inbody&&!inscript)&&(strlen(text)<UDM_MAXTEXTSIZE)){
				if(*text)strcat(text," ");
				strncat(text,tmp,UDM_MAXTEXTSIZE);
			}
			if((intitle)&&(strlen(title)<UDM_MAXTITLESIZE)){
				if(*title)strcat(title," ");
				strncat(title,tmp,UDM_MAXTITLESIZE-1);
				title[UDM_MAXDESCSIZE-1]=0;
			}
			if(index&&!inscript){
				ParseText(Indexer,CurSrv,tmp,separator,
					CurSrv->bodyweight*inbody+
					CurSrv->titleweight*intitle,1);
			}
			free(tmp);
			s=s+len; state=HTML_UNK;
			break;
		}
	}
	return(0);
}

/****************************************************************/
static int cmplang(const void *s1,const void *s2){
	return(((UDM_LANG*)s2)->count-((UDM_LANG*)s1)->count);
}


static void FreeDoc(UDM_DOCUMENT *Result){
	if(!Result)return;
	UDM_FREE(Result->url);
	UDM_FREE(Result->content_type);
	UDM_FREE(Result->last_modified);
	UDM_FREE(Result->title);
	UDM_FREE(Result->text);
	UDM_FREE(Result->last_index_time);
	UDM_FREE(Result->next_index_time);
	UDM_FREE(Result->keywords);
	UDM_FREE(Result->description);
	UDM_FREE(Result->crc);
	free(Result);
}


/*********************** 'UrlFile' stuff (for -f option) *******************/
__INDLIB__ void AddURLFile(char *name){
	UDM_FREE(url_file_name);
	if(name)
		url_file_name=strdup(name);
}

__INDLIB__ int URLFile(UDM_INDEXER *Indexer, int action){
FILE *url_file;
char str[UDMSTRSIZ]="";
char str1[UDMSTRSIZ]="";
int result;
UDM_URL myurl;

	/*  Read lines and clear urls*/
	/* we've already tested in main.c to make sure it can be read */

	if(!strcmp(url_file_name,"-"))
		url_file=stdin;
	else
		url_file=fopen(url_file_name,"r");

	while(fgets(str1,sizeof(str1),url_file)){
		char *end;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		switch(action){
		case UDM_URL_FILE_REINDEX:
			AddURLLimit(str);
			
			LOCK_TARGET(Indexer);
			have_targets=1;
			UNLOCK_TARGET(Indexer);

			result=MarkForReindex(Indexer);
			if(result!=IND_OK)return(result);
			ClearURLLimit();
			break;
		case UDM_URL_FILE_CLEAR:
			AddURLLimit(str);
			result=ClearDB(Indexer);
			if(result!=IND_OK)return(IND_ERROR);
			ClearURLLimit();
			break;
		case UDM_URL_FILE_INSERT:
			if(AddHref(Indexer,str,0,0,0)){
				LOCK_TARGET(Indexer);
				have_targets=1;
				UNLOCK_TARGET(Indexer);
			}
			break;
		case UDM_URL_FILE_PARSE:
			if(parseurl(&myurl,str)){
				return(IND_ERROR);
			}
			break;
		}
		str[0]=0;
	}
	if(url_file!=stdin)
		fclose(url_file);
	return(IND_OK);
}


/***************** Main function IndexNextURL *******************/


/* Init section for indexing */

static int StartUp(UDM_INDEXER * Indexer, int index_flags){

	CurURLNumber=0;
	have_targets=1;

	if(index_flags&UDM_FLAG_MARK){
		if(url_file_name)
			URLFile(Indexer,UDM_URL_FILE_REINDEX);

		MarkForReindex(Indexer);
		if(DBErrorCode(Indexer->db)){
			return(1);
		}
	}

	if((url_file_name) && (index_flags&UDM_FLAG_INSERT)) {
		URLFile(Indexer,UDM_URL_FILE_INSERT);
		if(DBErrorCode(Indexer->db)){
			return(1);
		}
	}

	LoadRobots(Indexer);
	if(DBErrorCode(Indexer->db)){
		return(1);
	}

	load_stop_list(Indexer->db);
	if(DBErrorCode(Indexer->db)){
		return(1);
	}
	
	if(index_flags&UDM_FLAG_SKIP_LOCKING)
		DBUseLock=0;
	else
		DBUseLock=1;

	return(0);
}

__INDLIB__ int IndexNextURL(UDM_INDEXER *Indexer,int index_flags){
char str[UDMSTRSIZ]="",str1[UDMSTRSIZ]="";
char text[2*UDM_MAXTEXTSIZE]="";      char text_escaped[2*UDM_MAXTEXTSIZE+1]="";
char keywords[UDM_MAXKEYWORDSIZE]=""; char keywords_escaped[2*UDM_MAXKEYWORDSIZE+1]="";
char descript[UDM_MAXDESCSIZE]="";    char descript_escaped[2*UDM_MAXDESCSIZE+1]="";
char last_modified[32]="";        char last_modified_escaped[32*2+1]="";
char title[UDM_MAXTITLESIZE]="";      char title_escaped[UDM_MAXTITLESIZE*2+1];
char subj[UDM_MAXTITLESIZE]="";
char from[UDM_MAXKEYWORDSIZE]="";	char digest[33];

#ifdef NEWS_EXT
// I need to retain newsgroup-info, so I introduce new variables
// I also stick with these fixed-length strings, although I dislike them.
char HeaderDate[UDM_MAXDATESIZE] = ""; 	char HeaderDateEsc[2*UDM_MAXDATESIZE+1] = "";
char HeaderSubject[UDM_MAXSUBJSIZE] = "";	char HeaderSubjectEsc[2*UDM_MAXSUBJSIZE+1] = "";
char HeaderFrom[UDM_MAXFROMSIZE] = "";	char HeaderFromEsc[2*UDM_MAXFROMSIZE+1] = "";
char HeaderGroup[UDM_MAXGROUPSIZE] = "";	char HeaderGroupEsc[2*UDM_MAXGROUPSIZE+1] = "";
char HeaderRefs[UDM_MAXREFSIZE] = "";	char HeaderRefsEsc[2*UDM_MAXREFSIZE+1] = "";
char MessageIdEsc[2*UDM_MAXFROMSIZE+1]="";
char *parent=NULL, *SQLDate;
#endif

char *s,*content=NULL,*header,*content_type,*location,*href,*statusline,*lt;
int j,size,status=0,realsize,Method=0;
int origin,clone,follow,index,changed;
int reindex;
int result=IND_UNKNOWN;
int found_in_mirror = 0;
char *hcopy=NULL;
int DCindex=0; /* Document charset index number */

#ifdef USE_PARSER
char *mime;
int mimeno;
#endif

UDM_URL curURL,newURL,realURL;
UDM_DOCUMENT * Doc;
UDM_FILTER *MatchedFilter=NULL;
UDM_SERVER *CurSrv=0;
UDM_ALIAS *Alias;

	if(index_flags&UDM_FLAG_INIT){
		if(StartUp(Indexer, index_flags))
			return(IND_ERROR);
		else
			return(IND_OK);
	}

	/* Store URLs from cache into database */
	if(StoreHrefs(Indexer)){
		LOCK_TARGET(Indexer);
		have_targets=1;
		UNLOCK_TARGET(Indexer);
	}
	if(DBErrorCode(Indexer->db))
		return(IND_ERROR);

#ifdef HAVE_PTHREAD
	LOCK_HREFS(Indexer);
	LOCK_TARGET(Indexer);
	j=have_targets;
	UNLOCK_TARGET(Indexer);
	UNLOCK_HREFS(Indexer);
	if(!j){
		return(IND_NO_TARGET);
	}
#endif


	if((GetIndexedNumber()>=MaxURLNumber)&&(MaxURLNumber>=0))
		return(IND_NO_TARGET);

	/* Get Next URL to be indexed from the database */
	THREADINFO(Indexer->handle,"Selecting","");
	if(!(Doc=GetDocInfo(Indexer,index_flags))){
		if(DBErrorCode(Indexer->db)){
			return(IND_ERROR);
		}else{
			LOCK_TARGET(Indexer);
			have_targets=0;
			UNLOCK_TARGET(Indexer);
			return(IND_NO_TARGET);
		}
	}
	if(Indexer->action==IND_TERMINATED)return(IND_TERMINATED);
	IncURLNumber();

	reindex=index_flags&UDM_FLAG_REINDEX;

	if(!Indexer->buf){
		/* Alloc buffer for document */
		Indexer->buf=(char*)malloc(MaxDocSize);
	}

	/* Clear language statistics */
	for(j=0;j<UDM_LANGPERDOC;j++){
		Indexer->lang[j].lang[0]=0;
		Indexer->lang[j].count=0;
	}

	udm_log(Indexer->handle,UDM_LOG_INFO,"%s",Doc->url);
	THREADINFO(Indexer->handle,"Found",Doc->url);

	/* Add current URL in cache to avoid */
	/* possible double INSERT INTO url   */
	if(AddHref(Indexer,Doc->url,0,Doc->hops,1)){
		LOCK_TARGET(Indexer);
		have_targets=1;
		UNLOCK_TARGET(Indexer);
	}


#ifndef HAVE_PTHREAD
#ifdef HAVE_SETPROCTITLE
	setproctitle("%s",Doc->url);
#endif
#endif

	/* Check that URL has valid syntax */
	if(parseurl(&curURL,Doc->url)){
		result=DeleteUrl(Indexer,Doc->url_id);
		udm_log(Indexer->handle,UDM_LOG_WARN,"Invalid URL: %s ... deleted",Doc->url);
		FreeDoc(Doc);
		return(result);
	}

	/* Find corresponded Server record from indexer.conf */
	if(!(CurSrv=FindServer(Doc->url))){
		if(!strcmp(curURL.filename,"robots.txt")){
			if(IND_OK==(result=DeleteRobotsFromHost(Indexer,curURL.hostinfo)))
				result=LoadRobots(Indexer);
		}else{
			udm_log(Indexer->handle,UDM_LOG_WARN,"No 'Server' command for url... deleted.");
			result=IND_OK;
		}
		if(result==IND_OK)result=DeleteUrl(Indexer,Doc->url_id);
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}

	/* Check that hops is less than MaxHops */
	if(Doc->hops>CurSrv->maxhops){
		result=DeleteUrl(Indexer,Doc->url_id);
		udm_log(Indexer->handle,UDM_LOG_WARN,"Too many hops (%d)... deleted.",Doc->hops);
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}

	/* Check Allow/Disallow/CheckOnly stuff */
	if(!(MatchedFilter=FindFilter(Doc->url)))
		Method=UDM_ALLOW;
	else
		Method=MatchedFilter->filter_type;

	if(Method==UDM_DISALLOW){
		udm_log(Indexer->handle,UDM_LOG_WARN,"Deleting URL: '%s'",FilterComStr(MatchedFilter));
		result=DeleteUrl(Indexer,Doc->url_id);
		if((result==IND_OK)&&(!strcmp(curURL.filename,"robots.txt"))){
			if(IND_OK==(result=DeleteRobotsFromHost(Indexer,curURL.hostinfo)))
				result=LoadRobots(Indexer);
		}
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}

	/* Check whether URL is disallowed by robots.txt */
	if((CurSrv->userobots)&&(FindRobots(Indexer,Doc->url)>=0)){
		udm_log(Indexer->handle,UDM_LOG_WARN,"Deleting URL: robots.txt");
		result=DeleteUrl(Indexer,Doc->url_id);
	}else
	/* Check whether FTP is used without proxy */
	if((!strcmp(curURL.schema,"ftp"))&&(!CurSrv->proxy)){
		udm_log(Indexer->handle,UDM_LOG_WARN,"FTP is available via proxy only, skipped");
		result=UpdateUrl(Indexer,Doc->url_id,504,CurSrv->period);
	}else
	/* Check that too many errors on this server */
	if(CurSrv->net_errors>CurSrv->max_net_errors){
		udm_log(Indexer->handle,UDM_LOG_WARN,"Too many network errors for this server, skipped");
		result=UpdateUrl(Indexer,Doc->url_id,504,CurSrv->period);
	}
	if(result){
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}

	/* Find alias */
	parseurl(&realURL,Doc->url);
	if((Alias=FindAlias(Doc->url))){
		sprintf(str1,"%s%s",Alias->replace,Doc->url+strlen(Alias->find));
		freeurl(&realURL);
		if(parseurl(&realURL,str1)){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"Error in aliased URL: '%s'",str1);
			freeurl(&curURL);
			FreeDoc(Doc);
			return(IND_OK);
		}else{
			udm_log(Indexer->handle,UDM_LOG_EXTRA,"Alias: '%s'",str1);
		}
	}


	/* Compose HTTP header */
	sprintf(str,"%s %s%s HTTP/1.0\r\n",
		(Method==UDM_HEAD)?"HEAD":"GET",
		CurSrv->proxy?Doc->url:realURL.path,
		CurSrv->proxy?"":realURL.filename);

	if((Doc->status)&&(last_modified[0])&&(!reindex))
		sprintf(UDM_STREND(str),"If-Modified-Since: %s\r\n",
			last_modified);

	if((!strcmp(realURL.schema,"http"))||(!strcmp(realURL.schema,"ftp"))){

		/* User agent */
		sprintf(UDM_STREND(str),"User-Agent: %s\r\n",UserAgent);

		/* If LocalCharset specified */
		if(LocalCharset)
			sprintf(UDM_STREND(str),"Accept-charset: %s\r\n",
				LocalCharset);
		/* Host Name for virtual hosts */
		sprintf(UDM_STREND(str),"Host: %s\r\n",realURL.hostname);

		/* Auth if required */
		if(CurSrv->basic_auth)
			sprintf(UDM_STREND(str),"Authorization: Basic %s\r\n",
				CurSrv->basic_auth);
	}

	/* Add user defined headers */
	if(*ExtraHeaders)strcat(str,ExtraHeaders);

	/* Empty line is the end of HTTP header */
	strcat(str,"\r\n");

	size=UDM_NET_UNKNOWN;

	THREADINFO(Indexer->handle,"Getting",Doc->url);

	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;

	if (use_mirror >= 0) {
		char errstr[UDMSTRSIZ]="";
		/* on u_m==0 it returned by mtime from mirrorget */
		/* but we knew that it should be put in mirror  */

		size = MirrorGET(realURL.schema, realURL.hostname,
				realURL.path, realURL.filename,
				Indexer->buf, MaxDocSize,use_mirror,errstr);
		if(size>0){
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s has been taken from mirror",Doc->url);
			found_in_mirror=1;
		}else{
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s",errstr);
		}
	}

	if(!found_in_mirror){
#ifdef USE_HTDB
		if(!strcmp(realURL.schema,"htdb")){
			size = HTDBGet(Indexer,realURL.path,realURL.filename,
				CurSrv->htdb_list,CurSrv->htdb_doc,
				Indexer->buf,MaxDocSize);
			if(DBErrorCode(Indexer->db))result=IND_ERROR;
		}
#endif
#ifdef USE_FILE
		if(!strcmp(realURL.schema,"file")){
			size = FILEGet(str,Indexer->buf,MaxDocSize);
		}
#endif
#ifdef USE_NEWS
		if((!strcmp(realURL.schema,"news"))){
			size = NEWSGet(realURL.hostname,realURL.port?realURL.port:119,
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}else
		if((!strcmp(realURL.schema,"nntp"))){
			size = NNTPGet(realURL.hostname,realURL.port?realURL.port:119,
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}
#endif
#ifdef USE_HTTP
		if((!strcmp(realURL.schema,"http"))){
			size = HTTPGet(
				CurSrv->proxy?CurSrv->proxy:realURL.hostname,
				CurSrv->proxy?CurSrv->proxy_port:(realURL.port?realURL.port:80),
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}else
		if((!strcmp(realURL.schema,"ftp"))){
			size = HTTPGet(
				CurSrv->proxy?CurSrv->proxy:realURL.hostname,
				CurSrv->proxy?CurSrv->proxy_port:(realURL.port?realURL.port:80),
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}
#endif
	}


	if(!result)
	switch(size){
	case UDM_NET_UNKNOWN:
		udm_log(Indexer->handle,UDM_LOG_WARN,"Protocol not supported");
		status=UDM_HTTP_STATUS_NOT_SUPPORTED;
		result=IND_OK;
		break;
	case UDM_NET_TIMEOUT:
		udm_log(Indexer->handle,UDM_LOG_WARN,"Download timeout");
		status=UDM_HTTP_STATUS_TIMEOUT;
		result=IND_OK;
		break;
	case UDM_NET_CANT_CONNECT:
		udm_log(Indexer->handle,UDM_LOG_WARN,"Can't connect to host %s:%d",
		CurSrv->proxy?CurSrv->proxy:curURL.hostname,
		CurSrv->proxy?CurSrv->proxy_port:(curURL.port?curURL.port:80));
		status=UDM_HTTP_STATUS_UNAVAIL;
		result=IND_OK;
		break;
	case UDM_NET_CANT_RESOLVE:
		udm_log(Indexer->handle,UDM_LOG_WARN,"Unknown host %s",
			CurSrv->proxy?CurSrv->proxy:curURL.hostname);
		status=UDM_HTTP_STATUS_UNAVAIL;
		result=IND_OK;
		break;
	default:
		if(size<0){	/* No connection */
			udm_log(Indexer->handle,UDM_LOG_WARN,"Can't connect to host %s:%d",
				CurSrv->proxy?CurSrv->proxy:curURL.hostname,
				CurSrv->proxy?CurSrv->proxy_port:(curURL.port?curURL.port:80));
			status=UDM_HTTP_STATUS_UNAVAIL;
			result=IND_OK;
		}else{
			/* Cut response header */
			if((content=strstr(Indexer->buf,"\r\n\r\n"))){
				*content = 0;
				content +=4;
			} else if ((content=strstr(Indexer->buf,"\n\n"))){
				*content = 0;
				content +=2;
			}
			if(!content){
				udm_log(Indexer->handle,UDM_LOG_ERROR,"Illegal HTTP headers in response");
				status=UDM_HTTP_STATUS_UNAVAIL;
				result=IND_OK;
			}
		}
	}
	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;
	if(result){
		CurSrv->net_errors++;
		freeurl(&curURL);
		freeurl(&realURL);
		if(result!=IND_ERROR)
			result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		FreeDoc(Doc);
		return(result);
	}

	Indexer->buf[size]=0;
	header=Indexer->buf;
	realsize=size;

	status=changed=1;clone=0;origin=0;
	content_type=location=href=Indexer->charset=statusline=NULL;
	digest[0]=0;text[0]=0;title[0]=0;keywords[0]=0;
	descript[0]=0;subj[0]=0;last_modified[0]=0;from[0]=0;
	strcpy(last_modified,Doc->last_modified?Doc->last_modified:"");

	index	=CurSrv->gindex;
	follow	=CurSrv->gfollow;

	size-=(content-Indexer->buf);     /* Could be modified by Content-Length    */
	realsize-=(content-Indexer->buf); /* Will be safe for DIGEST, Parsers etc   */

	THREADINFO(Indexer->handle,"Parsing",Doc->url);
	/* Now lets parse response header lines */
	hcopy = strdup(header);
	s=gettoken(header,"\r\n",&lt);
	while(s){
		if(!UDM_STRNCASECMP(s,"HTTP/")){
			status=atoi(s+8);
			statusline=s;
		}else
		if (Force1251 && !UDM_STRNCASECMP(s,"Server: ")){
			if (sscanf(s+7,"%s",str1)==1)
				if (!UDM_STRNCASECMP(str1,"Microsoft")||
					!UDM_STRNCASECMP(str1,"IIS"))
					Indexer->charset="windows-1251";
		}else
		if(!UDM_STRNCASECMP(s,"Content-Type: ")){
			char *p;
			content_type=s+14;
			if((p=strstr(content_type,"charset=")))
				Indexer->charset = p + 8;
		}else
		if(!UDM_STRNCASECMP(s,"Location: ")){
			location=s+10;
		}else
		if(!UDM_STRNCASECMP(s,"Content-Length: ")){
			size=atoi(s+16);
		}else
		if(!UDM_STRNCASECMP(s,"Subject: ")){
			UDM_STRNCPY(title,s+9);
			strcpy(subj,title);
#ifdef NEWS_EXT
			strncpy(HeaderSubject,s+9,UDM_MAXSUBJSIZE);
#endif
		}else
		if(!UDM_STRNCASECMP(s,"From: ")){
			UDM_STRNCPY(from,s+6);
#ifdef NEWS_EXT
			strncpy(HeaderFrom,s+6,UDM_MAXFROMSIZE);
#endif
		}else
		if(!UDM_STRNCASECMP(s,"Newsgroups: ")){
			UDM_STRNCPY(keywords,s+12);
#ifdef NEWS_EXT
			strncpy(HeaderGroup,s+12,UDM_MAXGROUPSIZE);
#endif
		}else
#ifdef NEWS_EXT
		if(!UDM_STRNCASECMP(s,"Date: ")){
			strncpy(HeaderDate,s+6,UDM_MAXDATESIZE);
		}else
		if(!UDM_STRNCASECMP(s,"References: ")){
			strncpy(HeaderRefs,s+12,UDM_MAXREFSIZE);
		}else
#endif
		if(!UDM_STRNCASECMP(s,"Last-Modified: ")){
			UDM_STRNCPY(last_modified,s+15);
			escstr(last_modified_escaped,last_modified);
		}
		udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s",s);
		s=gettoken(NULL,"\r\n",&lt);
	}
	udm_log(Indexer->handle,UDM_LOG_EXTRA,"%s %s %d",statusline?statusline:"?",content_type?content_type:"?",size);

#ifdef NEWS_EXT
	/* do correct string termination */
	HeaderDate[UDM_MAXDATESIZE-1] = 0;
	HeaderFrom[UDM_MAXFROMSIZE-1] = 0;
	HeaderSubject[UDM_MAXSUBJSIZE-1] = 0;
	HeaderGroup[UDM_MAXGROUPSIZE-1] = 0;
	HeaderRefs[UDM_MAXREFSIZE-1] = 0;
	if((SQLDate=parse_date(HeaderDate))){
		escstr(HeaderDateEsc,parse_date(HeaderDate));
		free(SQLDate);
	}
	escstr(HeaderFromEsc,HeaderFrom);
	escstr(HeaderSubjectEsc,HeaderSubject);
	escstr(HeaderGroupEsc,HeaderGroup);
	escstr(HeaderRefsEsc,HeaderRefs);
#endif

	switch(HTTPResponseType(status)){
	case 1: /* No HTTP code */
		CurSrv->net_errors++;
		udm_log(Indexer->handle,UDM_LOG_ERROR,"No HTTP response status");
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case UDM_HTTP_STATUS_OK:
		if(!content_type){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"No Content-type in '%s'!",Doc->url);
			CurSrv->net_errors++;
			result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		}
		break;

	case UDM_HTTP_STATUS_REDIRECT: /* We'll try to use Location: xxx instead */

		if((Doc->hops<CurSrv->maxhops)&&(location)){
			int newMethod;
			if((MatchedFilter=FindFilter(location)))
				newMethod=MatchedFilter->filter_type;
			else
				newMethod=UDM_ALLOW;
			if(newMethod!=UDM_DISALLOW){
				if(!parseurl(&newURL,location)){
					if(FindServer(location)){
						if((!CurSrv->delete_no_server)&&(!CurSrv->outside)){
							/* compare hostinfo in some cases */
							if(!strcmp(curURL.hostinfo,newURL.hostinfo)){
								if(AddHref(Indexer,location,Doc->url_id,Doc->hops+1,0)){
									LOCK_TARGET(Indexer);
									have_targets=1;
									UNLOCK_TARGET(Indexer);
								}
							}
						}else{
							if(AddHref(Indexer,location,Doc->url_id,Doc->hops+1,0)){
								LOCK_TARGET(Indexer);
								have_targets=1;
								UNLOCK_TARGET(Indexer);
							}
						}
					}
					freeurl(&newURL);
				}else{
					udm_log(Indexer->handle,UDM_LOG_ERROR,"Error in redirect URL: '%s'",location);
				}
			}
		}
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case UDM_HTTP_STATUS_NOT_MODIFIED:  /* Not Modified, nothing to do */
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case UDM_HTTP_STATUS_DELETE:
		/* delete it if not robots.txt */
		udm_log(Indexer->handle,UDM_LOG_EXTRA,"Deleting URL");
		if(!strcmp(curURL.filename,"robots.txt")){
			result=DeleteRobotsFromHost(Indexer,curURL.hostinfo);
			if(result==IND_OK)result=LoadRobots(Indexer);
		}
		if(result!=IND_ERROR){
			if(CurSrv->deletebad){
				result=DeleteUrl(Indexer,Doc->url_id);
			}else{
				result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
			}
		}
		break;

	case UDM_HTTP_STATUS_RETRY: /* We'll retry later, maybe host is down */
		CurSrv->net_errors++;
		udm_log(Indexer->handle,UDM_LOG_EXTRA,"Could not read URL");
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	default: /* Unknown status, retry later */
		udm_log(Indexer->handle,UDM_LOG_WARN,"HTTP %d We don't yet know how to handle it, skipped",status);
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
	}
	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;
	if(result){
		freeurl(&curURL);
		freeurl(&realURL);
		FreeDoc(Doc);
		return(result);
	}

	/* Now we have HTTP_OK and know Content-Type */

	if ((use_mirror>=0)&&(!found_in_mirror)){
		char errstr[UDMSTRSIZ]="";
		/* I tried to get it from mirror, but nothing was there  */
		/* or it was expired. Now we will store it on the mirror */
		if(MirrorPUT(realURL.schema, realURL.hostname,
				realURL.path, realURL.filename,
				hcopy, content, size,errstr)){
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s",errstr);
		}
	}
	UDM_FREE(hcopy);

#ifdef USE_PARSER
	/* Let's try to start external parser for this Content-Type */
	if((mime=mime_parse(content_type,&mimeno,content,realsize,MaxDocSize-(content-Indexer->buf)))){
		content=mime;
		content_type=mime_param(mimeno,UDM_MIME_TO);
		if((mime=mime_param(mimeno,UDM_MIME_CHARSET)))
			Indexer->charset=mime;
	};
#endif

	/* robots.txt */
	if(!UDM_STRNCASECMP(content_type,"text/plain")&&
	(!strcmp(curURL.filename,"robots.txt"))){
		result=ParseRobots(Indexer,content,curURL.hostinfo);
		if(result!=IND_ERROR)result=LoadRobots(Indexer);
		if(result==IND_ERROR){
			freeurl(&curURL);
			freeurl(&realURL);
			FreeDoc(Doc);
			return(IND_ERROR);
		}
	}else

	/* plain TEXT or the same */
	if(!UDM_STRNCASECMP(content_type,"text/plain")||
		!UDM_STRNCASECMP(content_type,"text/tab-separated-values")||
		!UDM_STRNCASECMP(content_type,"text/css"))
	{
		if(Indexer->charset)	{
			recode_string(content,getcharset(Indexer->charset),LCindex);
		} else {
#ifdef USE_CHARSET_GUESSER
			DCindex = guesscharset(content, getcharset(CurSrv->charset));
#else
			DCindex = getcharset(CurSrv->charset);
#endif
			recode_string(content,DCindex,LCindex);
		}
		if(Method!=UDM_HEAD){
			MD5_DIGEST(content,realsize,digest);
			changed=strcmp(digest,Doc->crc);
			origin=FindOrigin(Indexer,digest);
			clone=((origin)&&(origin!=Doc->url_id));
		}
		if(((index)&&(!clone)&&(changed))||reindex){
			strncpy(text,content,UDM_MAXTEXTSIZE-2);
			text[UDM_MAXTEXTSIZE-1]=0;
			ParseText(Indexer,CurSrv,content,SEPARATOR,CurSrv->bodyweight,1);
		}
	}else

	/* HTML text */
	if(!UDM_STRNCASECMP(content_type,"text/html")){

		if (Indexer->charset) {
			recode_string(content,getcharset(Indexer->charset),LCindex);
		} else {
#ifdef USE_CHARSET_GUESSER
			DCindex = guesscharset(content, getcharset(CurSrv->charset));
#else
			DCindex = getcharset(CurSrv->charset);
#endif
			recode_string(content,DCindex,LCindex);
		}
		if(Method!=UDM_HEAD){
			MD5_DIGEST(content,realsize,digest);
			changed=strcmp(digest,Doc->crc);
			origin=FindOrigin(Indexer,digest);
			clone=((origin)&&(origin!=Doc->url_id));
		}
		if(((index||follow)&&(!clone)&&(changed))||reindex){
			ParseHtml(Indexer,CurSrv,content,SEPARATOR,&curURL,
				Doc,index,follow,text,keywords,descript,title);
		}
	}else{
		/* Unknown Content-Type */
		index=0; /* What this line for? */
		if(Method!=UDM_HEAD){
			MD5_DIGEST(content,realsize,digest);
			changed=strcmp(digest,Doc->crc);
			origin=FindOrigin(Indexer,digest);
			clone=((origin)&&(origin!=Doc->url_id));
		}
	}
	
	if (strcmp(curURL.filename,"robots.txt")) {
	    if(CurSrv->urlweight) {
		strcpy(str,Doc->url);
		ParseText(Indexer,CurSrv,str,SEPARATOR,CurSrv->urlweight,1);
	    }	
	    else {
		if(CurSrv->urlhostweight) {
		    strcpy(str,curURL.hostname);
		    ParseText(Indexer,CurSrv,str,SEPARATOR,CurSrv->urlhostweight,1);
		}
		if(CurSrv->urlpathweight) {
		    strcpy(str,curURL.path);
		    ParseText(Indexer,CurSrv,str,SEPARATOR,CurSrv->urlpathweight,1);
		}
		if(CurSrv->urlfileweight) {
		    strcpy(str,curURL.filename);
		    ParseText(Indexer,CurSrv,str,SEPARATOR,CurSrv->urlfileweight,1);
		}
	    }	
	    
	}

	THREADINFO(Indexer->handle,"Updating",Doc->url);
	if(clone){
		udm_log(Indexer->handle,UDM_LOG_EXTRA,"Duplicate Document with #%d",origin);
		result=DeleteWordFromURL(Indexer,Doc->url_id);
		if(result==IND_OK)
			result=UpdateClone(Indexer,Doc->url_id,status,CurSrv->period,
				content_type,last_modified_escaped,digest);
	}else
	if((!changed)&&(!reindex)){
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
	}else{
		if(subj[0]){
			strcpy(str,subj);
			ParseText(Indexer,CurSrv,str,SEPARATOR,CurSrv->titleweight,1);
		}
		if((result=StoreWords(Indexer,Doc->url_id))==IND_OK){
			if(StoreHrefs(Indexer)){
				LOCK_TARGET(Indexer);
				have_targets=1;
				UNLOCK_TARGET(Indexer);
			}
			if(DBErrorCode(Indexer->db))
				result=IND_ERROR;
			else
				result=IND_OK;
		}
		if(result!=IND_ERROR){
			int total,i;
			char str[UDMSTRSIZ]="";
			char lang[3]="";
			
			/* Prepare text,keywords,description */
			if(!strcmp(realURL.schema,"news"))strcpy(keywords,from);

			s=text;
#ifndef NEWS_EXT
			while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
#endif
			text[UDM_MAXTEXTSIZE-1]=0;escstr(text_escaped,text);

			s=keywords;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			keywords[UDM_MAXKEYWORDSIZE-1]=0;escstr(keywords_escaped,keywords);

			s=descript;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			descript[UDM_MAXDESCSIZE-1]=0;escstr(descript_escaped,descript);

			s=title;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			title[UDM_MAXTITLESIZE-1]=0;escstr(title_escaped,title);

			qsort((void*)(Indexer->lang+1),UDM_LANGPERDOC-1,sizeof(UDM_LANG),cmplang);

			total=0;
			for(i=1;i<UDM_LANGPERDOC;i++)
				total+=Indexer->lang[i].count;

			if(total){
				sprintf(str,"total: %d",total);
				for(i=1;i<UDM_LANGPERDOC;i++){
					if(Indexer->lang[i].count)
						sprintf(UDM_STREND(str)," %s:%d (%d%%)",
						Indexer->lang[i].lang,
						Indexer->lang[i].count,
						(int)(Indexer->lang[i].count*100/total));
				}
				udm_log(Indexer->handle,UDM_LOG_DEBUG,"language: %s",str);
				if(Indexer->lang[1].count>10){
					strncpy(lang,Indexer->lang[1].lang,2);
					lang[2]=0;
				}
			}
#ifdef NEWS_EXT

			/*
			extract message id from url
			thanks guys, the message id is in realURL.filename !!
			valid message ids have an @ character
			*/

			if(strchr(realURL.filename,'@'))
				escstr(MessageIdEsc,realURL.filename);
						
			/* get rec_id from my parent out of db (if I have one...)*/

			if(strlen(HeaderRefs))
			{
				/* HeaderRefs contains all message ids of my predecessors, space separated*/
				/* my direct parent is the last in the list*/
				if((parent = strrchr(HeaderRefs,' ')))
				{	
					/* parent now points to the space character */
					/* skip it */
					++parent;
				}
				else
				{
					/* there is only one entry in HeaderRefs, so this is my parent */
					parent=HeaderRefs;	
				}	
			}
			result = 0;
			/* if the parent is really a message id, */
			/* get its rec_id from database          */
			if(parent && strlen(parent) && strchr(parent,'@'))
				result = FindMsgID(Indexer,parent);	
			/*
			now register me with my parent
			result is -1 if no parent was found
			*/
#ifdef HEIKODEBUG
			fprintf(stderr,"%s, me: %d, parent: %d\n",parent,Doc->url_id,result);
#endif
			if(result > 0)
				result = RegisterChild(Indexer,result,Doc->url_id);	


			result=LongUpdateUrl(Indexer,Doc->url_id,status,changed,size,CurSrv->period,
				CurSrv->hint,index,last_modified_escaped,
				text_escaped,title_escaped,
				content_type,keywords_escaped,
				descript_escaped,digest,lang,
				HeaderDateEsc,
				HeaderSubjectEsc,
				HeaderFromEsc,
				HeaderGroupEsc,
				HeaderRefsEsc,
				MessageIdEsc);
#else
			result=LongUpdateUrl(Indexer,Doc->url_id,status,changed,size,CurSrv->period,
				CurSrv->hint,index,last_modified_escaped,
				text_escaped,title_escaped,
				content_type,keywords_escaped,
				descript_escaped,digest,lang);
#endif			
		}
	}
	FreeWords(Indexer);
	freeurl(&curURL);
	freeurl(&realURL);
	FreeDoc(Doc);
	return(result);
}


/************ Misc functions *****************************/


__INDLIB__ UDM_INDEXER  * AllocIndexer(int handle){
UDM_INDEXER * result;
	result=(UDM_INDEXER*)malloc(sizeof(UDM_INDEXER));
	result->mwords=0;
	result->nwords=0;
	result->swords=0;
	result->handle=handle; /* Handle is used in multi-threaded version */
	result->action=0;
	result->state[0]=0;
	result->Word=NULL;
	result->buf=NULL;
	result->charset=NULL;
	result->db=AllocDB(UDM_OPEN_MODE_WRITE);
	return(result);
}

__INDLIB__ void FreeIndexer(UDM_INDEXER *Indexer){

	if(Indexer){
		FreeWords(Indexer);
		UDM_FREE(Indexer->Word);
		UDM_FREE(Indexer->buf);
		UDM_FREE(LocalCharset);
		FreeDB(Indexer->db);
		free(Indexer);
	}
}


__INDLIB__ int ClearDatabase(UDM_INDEXER * Indexer){
	return(ClearDB(Indexer));
}


__INDLIB__ char * IndexerErrorMsg(UDM_INDEXER *Indexer){
	return(DBErrorMsg(Indexer->db));
}


__INDLIB__ void SetAction(UDM_INDEXER *Indexer,int action){
	Indexer->action=action;
}


__INDLIB__ int ShowStatistics(UDM_INDEXER * Indexer){
int expired_total=0;
int total_total=0;
UDM_STAT *stat,*s;


#if (WIN32|WINNT)
#else
	printf("\n          UdmSearch statistics\n\n");
	printf("%10s %10s %10s\n","Status","Expired","Total");
	printf("   -----------------------------\n");
#endif
	stat=GetStatistics(Indexer);
	if(DBErrorCode(Indexer->db)){
		return(IND_ERROR);
	}
	if((s=stat)){
		while(stat->status!=-1){
			SHOWSTAT(Indexer->handle,stat->status,stat->expired,stat->total,
				HTTPErrMsg(stat->status));
			total_total+=stat->total;
			expired_total+=stat->expired;
			stat++;
		}
		free(s);
	}
#if (WIN32|WINNT)
		SHOWSTAT(Indexer->handle,-1,expired_total,total_total,"");
#else
	printf("   -----------------------------\n");
	printf("%10s %10d %10d","Total",expired_total,total_total);
	printf("\n\n");
#endif
	return(IND_OK);
}
